<?php

namespace Bluebirdday\TranssmartSmartConnect\Model\Carrier;

use Magento\Quote\Model\Quote\Address\RateRequest;
use Magento\Shipping\Model\Rate\Result;

/**
 * Class Delivery
 * @package Bluebirdday\TranssmartSmartConnect\Model\Carrier
 */
class Delivery extends \Magento\Shipping\Model\Carrier\AbstractCarrier implements \Magento\Shipping\Model\Carrier\CarrierInterface
{
    /**
     * {@inheritdoc}
     */
    protected $_code = 'transsmartdelivery';

    /**
     * Custom basic path for all SmartConnect Pickup method settings.
     *
     * @var string
     */
    private $baseConfigPath = 'transsmart/shipping_methods/';

    /**
     * @var \Magento\Framework\Json\Helper\Data
     */
    private $jsonHelper;

    /**
     * @var \Magento\Shipping\Model\Rate\ResultFactory
     */
    private $rateResultFactory;

    /**
     * @var \Magento\Quote\Model\Quote\Address\RateResult\MethodFactory
     */
    private $rateMethodFactory;

    /**
     * @var \Bluebirdday\TranssmartSmartConnect\Model\CarrierRepository
     */
    private $profileRepository;

    /**
     * @var \Bluebirdday\TranssmartSmartConnect\Model\Rate\Response\Parser
     */
    private $rateParser;

    /**
     * @var null|\Bluebirdday\TranssmartSmartConnect\Model\Booking\Profile[]
     */
    private $availableProfiles = null;

    /**
     * @var \Bluebirdday\TranssmartSmartConnect\Model\Booking\Profile\AllowedCountry
     */
    private $allowedCountry;

    /**
     * Delivery constructor.
     * @param \Magento\Framework\App\Config\ScopeConfigInterface $scopeConfig
     * @param \Magento\Quote\Model\Quote\Address\RateResult\ErrorFactory $rateErrorFactory
     * @param \Magento\Framework\Json\Helper\Data $jsonHelper
     * @param \Psr\Log\LoggerInterface $logger
     * @param \Magento\Shipping\Model\Rate\ResultFactory $rateResultFactory
     * @param \Magento\Quote\Model\Quote\Address\RateResult\MethodFactory $rateMethodFactory
     * @param \Bluebirdday\TranssmartSmartConnect\Model\Rate\Response\Parser $rateParser
     * @param \Bluebirdday\TranssmartSmartConnect\Model\Booking\ProfileRepository $profileRepository
     * @param \Bluebirdday\TranssmartSmartConnect\Model\Booking\Profile\AllowedCountry $allowedCountry
     * @param array $data
     */
    public function __construct(
        \Magento\Framework\App\Config\ScopeConfigInterface $scopeConfig,
        \Magento\Quote\Model\Quote\Address\RateResult\ErrorFactory $rateErrorFactory,
        \Magento\Framework\Json\Helper\Data $jsonHelper,
        \Psr\Log\LoggerInterface $logger,
        \Magento\Shipping\Model\Rate\ResultFactory $rateResultFactory,
        \Magento\Quote\Model\Quote\Address\RateResult\MethodFactory $rateMethodFactory,
        \Bluebirdday\TranssmartSmartConnect\Model\Rate\Response\Parser $rateParser,
        \Bluebirdday\TranssmartSmartConnect\Model\Booking\ProfileRepository $profileRepository,
        \Bluebirdday\TranssmartSmartConnect\Model\Booking\Profile\AllowedCountry $allowedCountry,
        array $data = []
    ) {
        $this->jsonHelper = $jsonHelper;
        $this->profileRepository = $profileRepository;
        $this->rateResultFactory = $rateResultFactory;
        $this->rateMethodFactory = $rateMethodFactory;
        $this->rateParser = $rateParser;
        $this->allowedCountry = $allowedCountry;
        parent::__construct($scopeConfig, $rateErrorFactory, $logger, $data);
    }

    /**
     * Overridden method because of custom configuration path
     *
     * @param string $field
     * @return mixed
     */
    public function getConfigData($field)
    {
        return $this->_scopeConfig->getValue(
            $this->baseConfigPath . $field,
            \Magento\Store\Model\ScopeInterface::SCOPE_STORE,
            $this->getStore()
        );
    }

    /**
     * @param RateRequest $request
     * @return bool|Result
     */
    public function collectRates(RateRequest $request)
    {
        /** @var \Magento\Shipping\Model\Rate\Result $result */
        $result = $this->rateResultFactory->create();

        $deliveryMethods = $this->getDisplayMethods();

        if (empty($deliveryMethods)) {
            return $deliveryMethods;
        }

        foreach ($deliveryMethods as $deliveryMethod) {
            if (!$deliveryMethod['is_active']
                || !$this->isProfileAvailable($deliveryMethod['carrier_profile'], $request->getDestCountryId())
            ) {
                continue;
            }
            /** @var \Magento\Quote\Model\Quote\Address\RateResult\Method $method */
            $method = $this->rateMethodFactory->create();

            $method->setCarrier($this->getCarrierCode());
            $method->setCarrierTitle($this->getConfigData('title'));

            $method->setMethod($this->getMethodName($deliveryMethod['carrier_profile']));
            $method->setMethodTitle($this->getMethodTitle($deliveryMethod));

            $method->setPrice((float) $deliveryMethod['price']);
            $method->setCost((float) $deliveryMethod['price']);

            $result->append($method);
        }

        return $result;
    }

    /**
     * @inheritdoc
     */
    public function getAllowedMethods()
    {
        return [$this->getCarrierCode() => $this->getConfigData('name')];
    }

    /**
     * @inheritdoc
     */
    public function isTrackingAvailable()
    {
        return true;
    }

    /**
     * Get method name.
     *
     * @param string $code
     * @return string
     */
    protected function getMethodName($code)
    {
        return sprintf('delivery_%s', $code);
    }

    /**
     * @param array $deliveryMethod
     * @return mixed
     */
    protected function getMethodTitle(array $deliveryMethod)
    {
        if (!empty($deliveryMethod['title_override'])) {
            return $deliveryMethod['title_override'];
        }
        $profiles = $this->getAvailableProfiles();

        return $profiles[$deliveryMethod['carrier_profile']]->getDescription();
    }

    /**
     * Check if current profile is available.
     *
     * @param string $profileCode
     * @param string $countryId two-letter country code
     * @return bool
     */
    private function isProfileAvailable($profileCode, $countryId)
    {
        $availableProfiles = $this->getAvailableProfiles();
        $availableProfiles = $this->allowedCountry->filterProfilesByCountry($availableProfiles, $countryId);
        return array_key_exists($profileCode, $availableProfiles);
    }

    /**
     * Get list of available profiles
     *
     * @return \Bluebirdday\TranssmartSmartConnect\Model\Booking\Profile[]|null
     */
    protected function getAvailableProfiles()
    {
        if (null === $this->availableProfiles) {
            $availableProfiles = [];
            foreach ($this->profileRepository->getList() as $profile) {
                $availableProfiles[$profile->getCode()] = $profile;
            }
            $this->availableProfiles = $availableProfiles;
        }

        return $this->availableProfiles;
    }

    /**
     * @return mixed
     */
    protected function getDisplayMethods()
    {
        return $this->jsonHelper->jsonDecode(
            $this->getConfigData('delivery_methods')
        );
    }
}
