<?php
namespace Bluebirdday\TranssmartSmartConnect\Model;

use Magento\Framework\App\Config\ScopeConfigInterface;
use Psr\Log\LoggerInterface;
use Bluebirdday\TranssmartSmartConnect\Model\ResourceModel\Costcenter\CollectionFactory as CostcenterCollectionFactory;

/**
 * Class CostcenterManagement
 * @package Bluebirdday\TranssmartSmartConnect\Model
 */
class CostcenterManagement
{
    const CONFIG_DEFAULT_PATH = 'transsmart/default_shipment_properties/costcenter';

    /**
     * @var CostcenterCollectionFactory
     */
    private $collectionFactory;

    /**
     * @var CostcenterRepository
     */
    private $costcenterRepository;

    /**
     * @var ScopeConfigInterface
     */
    private $scopeConfig;

    /**
     * @var LoggerInterface
     */
    private $logger;

    /**
     * @param CostcenterCollectionFactory $collectionFactory
     * @param CostcenterRepository $costcenterRepository
     * @param ScopeConfigInterface $scopeConfig
     * @param LoggerInterface $logger
     */
    public function __construct(
        CostcenterCollectionFactory $collectionFactory,
        CostcenterRepository $costcenterRepository,
        ScopeConfigInterface $scopeConfig,
        LoggerInterface $logger
    ) {
        $this->collectionFactory = $collectionFactory;
        $this->costcenterRepository = $costcenterRepository;
        $this->scopeConfig = $scopeConfig;
        $this->logger = $logger;
    }

    /**
     * Get default cost center with fallback: saved in config, default by Transsmart, first in the table
     *
     * @return \Bluebirdday\TranssmartSmartConnect\Model\Costcenter|null
     */
    public function getDefault()
    {
        $configDefault = $this->scopeConfig->getValue(static::CONFIG_DEFAULT_PATH, 'website');
        if ($configDefault === null) {
            $this->logger->error('Default const center is not configured. Please, re-save config.');
        } else {
            $defaultCostCenter = $this->costcenterRepository->load($configDefault);
            if ($defaultCostCenter->getId() == $configDefault) {
                return $defaultCostCenter;
            } else {
                $this->logger->error('Wrong default costcenter: ' . $configDefault);
            }
        }
        // if default is not configured or is deprecated, we fall back to Transsmart default
        $transsmartDefault = $this->getTranssmartDefault();
        if ($transsmartDefault) {
            return $transsmartDefault;
        }
        // panic mode: try to return first costcenter in the table if any
        return $this->getFirst();
    }

    /**
     * @return null|Costcenter
     */
    private function getTranssmartDefault()
    {
        /** @var \Bluebirdday\TranssmartSmartConnect\Model\ResourceModel\Costcenter\Collection $collection */
        $collection = $this->collectionFactory->create();
        $collection->addFieldToFilter('is_default', 1);
        $collection->getSelect()->limit(1);
        /** @var \Bluebirdday\TranssmartSmartConnect\Model\Costcenter $firstItem */
        if ($collection->count() === 0) {
            return null;
        }
        $firstItem = $collection->getFirstItem();
        return $firstItem;
    }

    /**
     * @return null|Costcenter
     */
    private function getFirst()
    {
        $collection = $this->collectionFactory->create();
        $collection->getSelect()->limit(1);
        if ($collection->count() === 0) {
            return null;
        }
        /** @var \Bluebirdday\TranssmartSmartConnect\Model\Costcenter $firstItem */
        $firstItem = $collection->getFirstItem();
        return $firstItem;
    }
}
