<?php
namespace Bluebirdday\TranssmartSmartConnect\Model\Entity;

use Magento\Framework\Exception\LocalizedException;

/**
 * Class Updater
 * Common entity updater for simple entities
 *
 * @package Bluebirdday\TranssmartSmartConnect\Model\Entity
 */
class Updater implements UpdaterInterface
{
    /**
     * @var \Bluebirdday\TranssmartSmartConnect\Model\Transsmart\Adapter
     */
    protected $adapter;

    /**
     * @var null|string
     */
    protected $adapterMethod;

    /**
     * @var RepositoryInterface
     */
    protected $repository;

    /**
     * @var \Magento\Framework\Data\Collection
     */
    protected $collection;

    /**
     * Updater constructor.
     * @param \Bluebirdday\TranssmartSmartConnect\Model\Transsmart\Adapter $adapter
     * @param RepositoryInterface $repository
     * @param \Magento\Framework\Data\Collection $collection
     * @param null|string $adapterMethod
     */
    public function __construct(
        \Bluebirdday\TranssmartSmartConnect\Model\Transsmart\Adapter $adapter,
        RepositoryInterface $repository,
        \Magento\Framework\Data\Collection $collection,
        $adapterMethod = null
    ) {
        $this->adapter = $adapter;
        $this->repository = $repository;
        $this->collection = $collection;
        $this->adapterMethod = $adapterMethod;
    }

    /**
     * Convert array to indexed array by code
     *
     * @param array $data
     * @return array
     */
    protected function setIndexByCode($data)
    {
        $result = [];
        foreach ($data as $item) {
            $result[$item->code] = $item;
        }
        return $result;
    }

    /**
     * Update existing models
     * Existing data objects are identified by code, saved and unset from initial $data
     * As a result returned $data contains only NEW data objects
     *
     * @param \stdClass[] $data
     * @return \stdClass[]
     */
    protected function updateExistingModels($data)
    {
        $existingModels = $this->repository->getList();

        // check all existing models: remove if deprecated or update if exists
        foreach ($existingModels as $existingModel) {
            $code = $existingModel->getCode();
            if (array_key_exists($code, $data)) {
                $dataObject = $data[$code];
                $this->fillModelData($existingModel, $dataObject);
                $this->repository->save($existingModel);
                //further we will save NEW data models, so we need to unset existing ones.
                unset($data[$code]);
            } else {
                $this->repository->delete($existingModel);
            }
        }

        return $data;
    }

    /**
     * Fill model with data from object
     *
     * @param \object $model
     * @param \stdClass $data
     */
    protected function fillModelData($model, $data)
    {
        $vars = get_object_vars($data);
        foreach ($vars as $name => $value) {
            $method = 'set'.ucfirst($name);
            if (method_exists($model, $method)) {
                $model->$method($value);
            } else {
                // currently we ignore extra fields
            }
        }
    }

    /**
     * Save data objects as models
     *
     * @param \stdClass[] $data
     */
    protected function saveNewModels($data)
    {
        foreach ($data as $dataItem) {
            $model = $this->collection->getNewEmptyItem();
            $this->fillModelData($model, $dataItem);
            $this->repository->save($model);
        }
    }

    public function execute()
    {
        if (method_exists($this->adapter, $this->adapterMethod)) {
            $data = call_user_func([$this->adapter, $this->adapterMethod]);
            $this->refreshEntityData($data);
        } else {
            throw new LocalizedException(__('Adapter method "%1" not found', $this->adapterMethod));
        }
    }

    /**
     * Refresh all carriers according to data received
     *
     * @param array $data
     */
    public function refreshEntityData($data)
    {
        //add references to entities here
        $data = $this->setIndexByCode($data);
        $filteredData = $this->updateExistingModels($data);
        $this->saveNewModels($filteredData);
    }
}
