<?php
namespace Bluebirdday\TranssmartSmartConnect\Model\Media;

use Magento\Framework\Exception\FileSystemException;
use Magento\Framework\Exception\LocalizedException;

class FileHandler
{
    /**
     * @var \Bluebirdday\TranssmartSmartConnect\Model\Media\Config
     */
    protected $mediaConfig;

    /**
     * @var \Magento\Framework\Filesystem\Directory\WriteInterface
     */
    protected $mediaDirectory;

    /**
     * @var \Magento\MediaStorage\Helper\File\Storage\Database
     */
    protected $fileStorageDb;

    /**
     * FileHandler constructor.
     * @param \Magento\Framework\Filesystem $filesystem
     * @param Config $mediaConfig
     * @param \Magento\MediaStorage\Helper\File\Storage\Database $fileStorageDb
     */
    public function __construct(
        \Magento\Framework\Filesystem $filesystem,
        \Bluebirdday\TranssmartSmartConnect\Model\Media\Config $mediaConfig,
        \Magento\MediaStorage\Helper\File\Storage\Database $fileStorageDb
    ) {
        $this->mediaDirectory = $filesystem->getDirectoryWrite(
            \Magento\Framework\App\Filesystem\DirectoryList::MEDIA
        );
        $this->mediaConfig = $mediaConfig;
        $this->fileStorageDb = $fileStorageDb;
    }

    /**
     * Get file extension by image data
     *
     * @param string $imageData
     * @return string
     * @throws FileSystemException
     */
    private function getExtension($imageData)
    {
        $details = [];
        $info = getimagesizefromstring($imageData, $details);
        // mime contains something like "image/png"
        $data = explode('/', $info['mime']);
        if ('image' !== array_shift($data)) {
            throw new FileSystemException(__('Image format not recognized'));
        }
        return array_shift($data);

    }

    /**
     * Get media file URL by name
     *
     * Wrapper for controller
     * @param string $fileName
     * @return string
     */
    public function getMediaUrl($fileName)
    {
        return $this->mediaConfig->getMediaUrl($fileName);
    }

    /**
     * Save media file
     *
     * @param string $fileName
     * @param $imageData
     * @return string
     * @throws FileSystemException
     * @throws LocalizedException
     */
    public function save($fileName, $imageData)
    {
        $destinationFileName = $fileName . '.' . $this->getExtension($imageData);
        $relativePath =  $this->mediaConfig->getMediaPath($destinationFileName);

        if ($this->fileStorageDb->checkDbUsage()) {
            $fileData = [
                'filename' => $destinationFileName,
                'directory' => $this->mediaConfig->getBaseMediaPath(),
                'content' => $imageData
            ];
            $result = $this->fileStorageDb->getStorageFileModel()->saveFile($fileData, false);
            if ($result) {
                $this->fileStorageDb->saveFile($relativePath);
            }
        } else {
            // we don't overwrite existing files
            if ($this->mediaDirectory->isExist($relativePath)) {
                return $destinationFileName;
            }
            $directory = $this->mediaConfig->getBaseMediaPath();
            if (!$this->mediaDirectory->create($directory)) {
                throw new FileSystemException(__('Cannot create directory ' . $directory));
            };
            if (!$this->mediaDirectory->isWritable($directory)) {
                throw new FileSystemException(__('Cannot write to directory ' . $directory));
            }
            $this->mediaDirectory->writeFile($relativePath, $imageData);
        }

        return $destinationFileName;
    }
}
