<?php
namespace Bluebirdday\TranssmartSmartConnect\Model\Rate\Data;
use Magento\Framework\Exception\LocalizedException;

/**
 * Class Collector
 * Collects data from customer quote and system config and structures it to array
 */
class Collector
{
    /**
     * @var \Magento\Checkout\Model\Session
     */
    private $checkoutSession;

    /**
     * @var \Magento\Quote\Model\QuoteRepository
     */
    private $quoteRepository;

    /**
     * @var \Magento\Framework\App\Config\ScopeConfigInterface
     */
    private $scopeConfig;

    /**
     * @var \Bluebirdday\TranssmartSmartConnect\Model\Data\AddressFactory
     */
    private $addressFactory;

    /**
     * @var \Bluebirdday\TranssmartSmartConnect\Model\Data\MeasurementsFactory
     */
    private $measurementFactory;

    /**
     * @var \Bluebirdday\TranssmartSmartConnect\Model\Data\RateFactory
     */
    private $rateFactory;

    /**
     * @var \Bluebirdday\TranssmartSmartConnect\Model\PackageManagement
     */
    private $packageManagement;

    /**
     * @var \Bluebirdday\TranssmartSmartConnect\Model\Data\PackageFactory
     */
    private $packageFactory;

    /**
     * @var \Bluebirdday\TranssmartSmartConnect\Model\Street\Address\Parser
     */
    private $streetParser;

    /**
     * @var \Bluebirdday\TranssmartSmartConnect\Model\Quote\AddressManager
     */
    private $addressManager;

    /**
     * Collector constructor.
     * @param \Magento\Checkout\Model\Session $checkoutSession
     * @param \Magento\Quote\Model\QuoteRepository $quoteRepository
     * @param \Magento\Framework\App\Config\ScopeConfigInterface $scopeConfig
     * @param \Bluebirdday\TranssmartSmartConnect\Model\Data\AddressFactory $addressFactory
     * @param \Bluebirdday\TranssmartSmartConnect\Model\Data\MeasurementsFactory $measurementFactory
     * @param \Bluebirdday\TranssmartSmartConnect\Model\Data\RateFactory $rateFactory
     * @param \Bluebirdday\TranssmartSmartConnect\Model\PackageManagement $packageManagement
     * @param \Bluebirdday\TranssmartSmartConnect\Model\Data\PackageFactory $packageFactory
     * @param \Bluebirdday\TranssmartSmartConnect\Model\Street\Address\Parser $streetParser
     * @param \Bluebirdday\TranssmartSmartConnect\Model\Quote\AddressManager $addressManager
     */
    public function __construct(
        \Magento\Checkout\Model\Session $checkoutSession,
        \Magento\Quote\Model\QuoteRepository $quoteRepository,
        \Magento\Framework\App\Config\ScopeConfigInterface $scopeConfig,
        \Bluebirdday\TranssmartSmartConnect\Model\Data\AddressFactory $addressFactory,
        \Bluebirdday\TranssmartSmartConnect\Model\Data\MeasurementsFactory $measurementFactory,
        \Bluebirdday\TranssmartSmartConnect\Model\Data\RateFactory $rateFactory,
        \Bluebirdday\TranssmartSmartConnect\Model\PackageManagement $packageManagement,
        \Bluebirdday\TranssmartSmartConnect\Model\Data\PackageFactory $packageFactory,
        \Bluebirdday\TranssmartSmartConnect\Model\Street\Address\Parser $streetParser,
        \Bluebirdday\TranssmartSmartConnect\Model\Quote\AddressManager $addressManager
    ) {
        $this->checkoutSession = $checkoutSession;
        $this->quoteRepository = $quoteRepository;
        $this->scopeConfig = $scopeConfig;
        $this->addressFactory = $addressFactory;
        $this->measurementFactory = $measurementFactory;
        $this->rateFactory = $rateFactory;
        $this->packageManagement = $packageManagement;
        $this->packageFactory = $packageFactory;
        $this->streetParser = $streetParser;
        $this->addressManager = $addressManager;
    }

    /**
     * Get store address field from config
     *
     * @param string $fieldName
     * @param bool $required
     * @return string
     * @throws LocalizedException
     */
    private function getStoreAddressField($fieldName, $required) {
        $billingFields = ['name', 'phone', 'merchant_vat_number'];

        $basicPath = in_array($fieldName, $billingFields) ? 'general/store_information/' : 'shipping/origin/';

        $fieldValue = $this->scopeConfig->getValue($basicPath . $fieldName);
        if ($required && empty($fieldValue)) {
            throw new LocalizedException(__('Field "%1" is not configured for the store', $fieldName));
        }
        return $fieldValue;
    }

    /**
     * Get all street lines of store address and parse them
     *
     * @return string[]
     */
    private function parseSenderStreet()
    {
        $storeStreets = [];
        $storeStreet = $this->getStoreAddressField('street_line1', true);
        $storeStreets[] = $storeStreet;

        $storeStreet2 = $this->getStoreAddressField('street_line2', false);
        if (!empty($storeStreet2)) {
            $storeStreets[] = $storeStreet2;
        }

        return $this->streetParser->parseStreetAddress($storeStreets);
    }

    /**
     * Get store address from system config and pack it to SmartConnect data structure
     *
     * TODO: replace with \Bluebirdday\TranssmartSmartConnectExtension\Model\Shipment\Creator::getSenderData
     * @return \Bluebirdday\TranssmartSmartConnect\Model\Data\Address
     */
    private function getSenderData()
    {
        $storeFields = [
            'city'    => ['city', true],
            'zipCode' => ['postcode', true],
            'country' => ['country_id', true],
        ];

        $senderAddressData = [
            'type' => 'SEND',
            'name' => 'Store owner',
            'residential' => 0,
        ];

        foreach ($storeFields as $transsmartName => $fieldData) {
            list($magentoName, $isRequired) = $fieldData;
            $senderAddressData[$transsmartName] = $this->getStoreAddressField($magentoName, $isRequired);
        }

        list($street, $houseNumber) = $this->parseSenderStreet();
        $senderAddressData['addressLine1'] = $street;
        $senderAddressData['houseNo'] = $houseNumber;
        return $this->addressFactory->create(['data' => $senderAddressData]);
    }

    /**
     * Parse receiver address from quote shipping address data
     *
     * @param \Magento\Quote\Model\Quote\Address $quoteShippingAddress
     * @return \Bluebirdday\TranssmartSmartConnect\Model\Data\Address
     */
    private function getReceiverAddress($quoteShippingAddress)
    {
        list($street, $houseNumber) = $this->streetParser->parseStreetAddress($quoteShippingAddress->getStreet());
        $receiverAddressData = [
            'type' => 'RECV',
            'name' => $quoteShippingAddress->getName(),
            'addressLine1' => $street,
            'houseNo' => $houseNumber,
            'city' => $quoteShippingAddress->getCity(),
            'zipCode' => $quoteShippingAddress->getPostcode(),
            'country' => $quoteShippingAddress->getCountryModel()->getData('iso2_code'),
            'residential' => $quoteShippingAddress->getCompany() ? 0 : 1,
        ];
        return $this->addressFactory->create(['data' => $receiverAddressData]);
    }

    /**
     * Set package measurement to SmartConnect Rate request Structure
     *
     * @param \Bluebirdday\TranssmartSmartConnect\Model\Package $package
     * @param \Magento\Quote\Model\Quote\Address\RateRequest|null $rateRequest
     * @return \Bluebirdday\TranssmartSmartConnect\Model\Data\Measurements
     */
    public function getPackageMeasurements($package, $rateRequest = null)
    {
        // let's check a source for weight info
        if ($rateRequest && $rateRequest->getPackageWeight()) {
            $packageWeight = $rateRequest->getPackageWeight();
            $weightUnit = $this->scopeConfig->getValue('general/locale/weight_unit');
        } else {
            $packageWeight = $package->getWeight();
            $weightUnit = $package->getMassUom();
        }

        // let's check a source for package size info
        if ($rateRequest && $rateRequest->getPackageHeight()
            && $rateRequest->getPackageWidth() && $rateRequest->getPackageDepth()
        ) {
            $packageHeight = $rateRequest->getPackageHeight();
            $packageWidth = $rateRequest->getPackageWidth();
            $packageLength = $rateRequest->getPackageDepth();
        } else {
            $packageHeight = $package->getHeight();
            $packageWidth = $package->getWidth();
            $packageLength = $package->getLength();
        }

        $packageParams = [
            'length' => $packageLength,
            'width' => $packageWidth,
            'height' => $packageHeight,
            'weight' => $packageWeight,
            'linearUom' => $package->getLinearUom(),
            'massUom' => $weightUnit,
        ];
        return $this->measurementFactory->create(['data' => $packageParams]);
    }

    /**
     * @param \Magento\Quote\Model\Quote\Address\RateRequest|null $rateRequest
     * @param \Bluebirdday\TranssmartSmartConnect\Model\Data\Measurements|string $measurements
     * @param string $packageType
     * @return mixed
     */
    private function getPackageData($rateRequest, $measurements, $packageType)
    {
        $packageData = [
            'lineNo' => 1,
            'packageType' => $packageType,
            'quantity' => $rateRequest ? $rateRequest->getPackageQty() : 1,
            'stackable' => false,
            'stackheight' => 1,
            'measurements' => $measurements,
        ];
        return $this->packageFactory->create(['data' => $packageData]);
    }

    /**
     * Collect data from different places to create shipping rate request to SmartConnect
     *
     * @param \Magento\Quote\Model\Quote\Address\RateRequest $rateRequest
     * @return \Bluebirdday\TranssmartSmartConnect\Model\Data\Rate
     * @throws LocalizedException
     */
    public function prepareRateRequest($rateRequest)
    {
        $quoteId = $this->checkoutSession->getQuoteId();
        $quote = $this->quoteRepository->getActive($quoteId);

        $senderAddress = $this->getSenderData();
        $estimationAddress = $this->addressManager->getEstimationAddress($quote);
        if ($estimationAddress) {
            $receiverAddress = $this->getReceiverAddress($estimationAddress);
        } else {
            $receiverAddress = $this->getReceiverAddress($quote->getShippingAddress());
        }
//        [$rateRequest->getDestStreet()]
//        $rateRequest->getDestCity()
//        $rateRequest->getDestPostcode()
//        $rateRequest->getDestCountryId() // two-letter country code

        //fill default box params
        $defaultPackage = $this->packageManagement->getDefault();
        if ($defaultPackage) {
            $measurements = $this->getPackageMeasurements($defaultPackage, $rateRequest);
            $packageType = $defaultPackage->getType();
        } else {
            throw new LocalizedException(__('Default package type not found'));
        }

        $package = $this->getPackageData($rateRequest, $measurements, $packageType);
        $rateParams = [
            'reference' => (string)crc32($quote->getCustomerFirstname() . ' ' . $quote->getCustomerLastname()),
            'costcenter' => '',
            'value' => 0,
            'measurements' => $measurements,
            'addresses' => [$senderAddress, $receiverAddress],
            'packages' => [$package],
            'language' => '',
            'currency' => $this->scopeConfig->getValue('currency/options/default'),
            'service' => '',
            'serviceLevelTime' => '',
            'serviceLevelOther' => '',
            'inbound' => 0,
            'loadmeters' => 0,
        ];
        $rate = $this->rateFactory->create(['data' => $rateParams]);
        return $rate;
    }

    public function prepareFromAddress()
    {
        $quoteId = $this->checkoutSession->getQuoteId();
        $quote = $this->quoteRepository->getActive($quoteId);

        $senderAddress = $this->getSenderData();
        if ($this->addressManager->getEstimationAddress($quote)) {
            $receiverAddress = $this->getReceiverAddress($this->addressManager->getEstimationAddress($quote));
        } else {
            $receiverAddress = $this->getReceiverAddress($quote->getShippingAddress());
        }

        //fill default box params
        $defaultPackage = $this->packageManagement->getDefault();
        if ($defaultPackage) {
            $measurements = $this->getPackageMeasurements($defaultPackage);
        } else {
            throw new LocalizedException(__('Cannot request rates. Default package is not configured.'));
        }

        $package = $this->getPackageData(null, $measurements, $defaultPackage->getType());
        $rateParams = [
            'reference' => (string)crc32($quote->getCustomerFirstname() . ' ' . $quote->getCustomerLastname()),
            'costcenter' => '',
            'value' => 0,
            'measurements' => $measurements,
            'addresses' => [$senderAddress, $receiverAddress],
            'packages' => [$package],
            'language' => '',
            'currency' => $this->scopeConfig->getValue('currency/options/default'),
            'service' => '',
            'serviceLevelTime' => '',
            'serviceLevelOther' => '',
            'inbound' => 0,
            'loadmeters' => 0,
        ];
        $rate = $this->rateFactory->create(['data' => $rateParams]);
        return $rate;
    }
}
