<?php
namespace Bluebirdday\TranssmartSmartConnect\Model\Rate\Response;

use Magento\Framework\Exception\LocalizedException;

class Parser
{
    /**
     * @var \Bluebirdday\TranssmartSmartConnect\Model\Transsmart\Adapter
     */
    private $apiAdapter;

    /**
     * @var \Bluebirdday\TranssmartSmartConnect\Model\Rate\Data\Collector
     */
    private $dataCollector;

    /**
     * @var \Bluebirdday\TranssmartSmartConnect\Model\Booking\ProfileRepository
     */
    private $profileRepository;

    /**
     * @var \Magento\Framework\App\Config\ScopeConfigInterface
     */
    private $scopeConfig;

    /**
     * @var \Bluebirdday\TranssmartSmartConnect\Model\Booking\Profile\AllowedCountry
     */
    private $allowedCountry;

    /**
     * Parser constructor.
     * @param \Bluebirdday\TranssmartSmartConnect\Model\Transsmart\Adapter $apiAdapter
     * @param \Bluebirdday\TranssmartSmartConnect\Model\Rate\Data\Collector $dataCollector
     * @param \Bluebirdday\TranssmartSmartConnect\Model\Booking\ProfileRepository $profileRepository
     * @param \Magento\Framework\App\Config\ScopeConfigInterface $scopeConfig
     * @param \Bluebirdday\TranssmartSmartConnect\Model\Booking\Profile\AllowedCountry $allowedCountry
     */
    public function __construct(
        \Bluebirdday\TranssmartSmartConnect\Model\Transsmart\Adapter $apiAdapter,
        \Bluebirdday\TranssmartSmartConnect\Model\Rate\Data\Collector $dataCollector,
        \Bluebirdday\TranssmartSmartConnect\Model\Booking\ProfileRepository $profileRepository,
        \Magento\Framework\App\Config\ScopeConfigInterface $scopeConfig,
        \Bluebirdday\TranssmartSmartConnect\Model\Booking\Profile\AllowedCountry $allowedCountry
    ) {
        $this->apiAdapter = $apiAdapter;
        $this->dataCollector = $dataCollector;
        $this->profileRepository = $profileRepository;
        $this->scopeConfig = $scopeConfig;
        $this->allowedCountry = $allowedCountry;
    }

    /**
     * Update request with data from pickupAddress
     *
     * @param \Magento\Quote\Model\Quote\Address\RateRequest $request
     * @param \Magento\Quote\Model\Quote\Address $pickupAddress
     */
    public function updateRequestInfo($request, $pickupAddress)
    {
        $requestFields = $request->getData();
        $addressFields = $pickupAddress->getData();
        foreach ($addressFields as $name => $value) {
            $updatedName = 'dest_' . $name;
            if (array_key_exists($updatedName, $requestFields)) {
                $request->setData($updatedName, $value);
            }
        }
    }

    /**
     * Rebuild profiles array with special indexes for quick comparison with rates
     *
     * @param \Bluebirdday\TranssmartSmartConnect\Model\Booking\Profile[] $profiles
     * @param string $delimiter
     * @return \Bluebirdday\TranssmartSmartConnect\Model\Booking\Profile[]
     */
    public function indexProfiles($profiles, $delimiter)
    {
        $indexedProfiles = [];
        foreach ($profiles as $profile) {
            $index = implode(
                $delimiter,
                [
                    $profile->getCarrierCode(),
                    $profile->getServiceLevelTimeCode(),
                    $profile->getServiceLevelOtherCode()
                ]
            );
            $indexedProfiles[$index] = $profile;
        }
        return $indexedProfiles;
    }

    /**
     * Get active profiles info filtered by country availability
     *
     * @param string $countryId
     * @return \stdClass[]
     * @throws LocalizedException
     */
    public function getPickupProfiles($countryId)
    {
        $availableProfiles = $this->scopeConfig->getValue('transsmart/pickup/profiles');
        if (!$availableProfiles) {
            throw new LocalizedException(__('No profiles configured for pickup'), null, 400);
        }
        $availableProfiles = json_decode($availableProfiles);
        if (!$availableProfiles && !is_array($availableProfiles)) {
            throw new LocalizedException(__('Cannot decode pickup profiles'), null, 400);
        }
        $activeProfiles = $this->filterActiveProfiles($availableProfiles);

        // we load profile objects and filter them by country
        $profileObjects = $this->profileRepository->getByCodes(array_keys($activeProfiles));
        $profileObjects = $this->allowedCountry->filterProfilesByCountry($profileObjects, $countryId);
        $activeProfiles = array_intersect_key($activeProfiles, $profileObjects);

        return $activeProfiles;
    }

    /**
     * @param \stdClass $configData
     * @return \stdClass[]
     */
    private function filterActiveProfiles($configData)
    {
        $activeProfiles = [];

        foreach ($configData as $config) {
            if ($config->is_active) {
                $profileData = new \stdClass();
                $profileData->description = $config->title_override;
                $profileData->salesPrice = floatval($config->price);
                $activeProfiles[$config->carrier_profile] = $profileData;
            }
        }

        return $activeProfiles;
    }

    /**
     * Get rates configured on Magento backend
     *
     * @param \stdClass[] $activeProfiles
     * @return array
     * @throws LocalizedException
     */
    public function getFixedRates($activeProfiles)
    {
        $profiles = $this->profileRepository->getByCodes(array_keys($activeProfiles));

        $result = [];
        foreach ($profiles as $profile) {
            $profileRate = $activeProfiles[$profile->getCode()];
            $profileRate->currency = $this->scopeConfig->getValue('currency/options/default');
            $profileRate->carrier = $profile->getCarrierCode();
            $profileRate->profile = $profile->getCode();
            if (!$activeProfiles[$profile->getCode()]->description) {
                $profileRate->description = $profile->getDescription();
            }
            $result[$profile->getCarrierCode()][] = $profileRate;
        }
        return $result;
    }

    /**
     * Filter rates which are equivalent to allowed profiles
     *
     * Filtering is performed by 'carrier code', 'service level time' and 'service level other' attributes
     *
     * @param \stdClass[] $rates
     * @param array $carrierProfileConfigs carrier profile configs
     * @return array
     */
    private function filterRates($rates, $carrierProfileConfigs)
    {
        $delimiter = '::';
        $profileCodes = array_keys($carrierProfileConfigs);
        $profiles = $this->profileRepository->getByCodes($profileCodes);
        $profiles = $this->indexProfiles($profiles, $delimiter);

        $allowedRates = [];
        foreach ($rates as $rate) {
            //we recognize profile by combination of carrier, service level time and service level other fields
            $rateKey = implode(
                $delimiter,
                [
                    $rate->carrier,
                    $rate->serviceLevelTime,
                    $rate->serviceLevelOther
                ]
            );
            if (array_key_exists($rateKey, $profiles)) {
                $rate->profile = $profiles[$rateKey]->getCode();
                $allowedRates[] = $rate;
            }
        }

        return $allowedRates;
    }

    /**
     * Get minimal rate per each carrier
     *
     * @param \stdClass[] $rates
     * @param array $carrierProfiles
     * @return array
     */
    public function getCarrierRates($rates, $carrierProfiles)
    {
        $rates = $this->filterRates($rates, $carrierProfiles);

        //group by carrier
        $carrierRates = [];
        foreach ($rates as $rate) {
            $carrierRates[$rate->carrier][] = $rate;
        }

        return $carrierRates;
    }

    /**
     * Get minimal 'salesPrice' attribute value among the rates
     *
     * @param $rates
     * @return null|float
     */
    public function getMinimumRate($rates)
    {
        if (empty($rates)) {
            return null;
        }

        $min = $rates[0]->salesPrice;
        foreach ($rates as $rate) {
            if ($rate->salesPrice < $min) {
                $min = $rate->salesPrice;
            }
        }
        return $min;

    }

    /**
     * @param \Magento\Quote\Model\Quote\Address\RateRequest $rateRequest
     * @param \Bluebirdday\TranssmartSmartConnect\Model\Booking\Profile|null $profile
     * @return float|null
     * @throws LocalizedException
     */
    public function getShippingAmount($rateRequest, $profile = null)
    {
        if (!$this->scopeConfig->getValue('transsmart/pickup/dynamic_rates')) {
            if ($profile) {
                $errorMessage = __('Preconfigured profile "%1" not found', $profile->getCode());
                $pickupProfiles = $this->getPickupProfiles($rateRequest->getDestCountryId());
                if (!array_key_exists($profile->getCode(), $pickupProfiles)) {
                    throw new LocalizedException($errorMessage);
                }
                $rates = $this->getFixedRates([$profile->getCode() => $pickupProfiles[$profile->getCode()]]);
            } else {
                $profiles = $this->getPickupProfiles($rateRequest->getDestCountryId());
                $rates = $this->getFixedRates($profiles);
                $errorMessage = __('No active pickup profiles found');
            }
            if (empty($rates)) {
                throw new LocalizedException($errorMessage);
            }
            $oneCarrierRates = array_shift($rates);
            return $this->getMinimumRate($oneCarrierRates);
        }

        $transsmartRequest = $this->dataCollector->prepareRateRequest($rateRequest);
        /** @var \stdClass[] $rates */
        $rates = $this->apiAdapter->getRates($transsmartRequest);
        if ($profile instanceof \Bluebirdday\TranssmartSmartConnect\Model\Booking\Profile) {
            $rates = $this->filterRates($rates, [$profile->getCode() => $profile->getCode()]);
            if (empty($rates)) {
                throw new \RuntimeException(__('Shipping rate for profile "%1" not found'), $profile->getDescription());
            }
            $firstRate = reset($rates);
            return $firstRate->salesPrice;
        } else {
            return $this->getMinimumRate($rates);
        }
    }
}
