<?php
namespace Bluebirdday\TranssmartSmartConnect\Model\Shipping\Method;

use Magento\Framework\Exception\LocalizedException;

class Parser
{
    /**
     * @var \Bluebirdday\TranssmartSmartConnect\Model\Booking\ProfileRepository
     */
    private $profileRepository;

    /**
     * @var \Magento\Framework\App\Config\ScopeConfigInterface
     */
    private $scopeConfig;

    /**
     * Parser constructor.
     * @param \Bluebirdday\TranssmartSmartConnect\Model\Booking\ProfileRepository $profileRepository
     * @param \Magento\Framework\App\Config\ScopeConfigInterface $scopeConfig
     */
    public function __construct(
        \Bluebirdday\TranssmartSmartConnect\Model\Booking\ProfileRepository $profileRepository,
        \Magento\Framework\App\Config\ScopeConfigInterface $scopeConfig
    ) {
        $this->profileRepository = $profileRepository;
        $this->scopeConfig = $scopeConfig;
    }

    /**
     * @param string $shippingMethod
     * @return bool
     */
    public function isPickup($shippingMethod)
    {
        $shippingMethodParts = explode('_', $shippingMethod);
        return $shippingMethodParts[0] === 'transsmartpickup' && count($shippingMethodParts) === 2;
    }

    /**
     * @param string $shippingMethod
     * @return bool
     */
    public function isDeliveryOption($shippingMethod)
    {
        $shippingMethodParts = explode('_', $shippingMethod);
        return $shippingMethodParts[0] === 'transsmartdelivery' && count($shippingMethodParts) === 3;
    }

    /**
     * @param string $shippingMethod
     * @return array
     * @throws LocalizedException
     */
    public function getPickupCarrierAndProfile($shippingMethod)
    {
        $shippingMethodParts = explode('_', $shippingMethod);
        if ($shippingMethodParts[0] === 'transsmartpickup' && count($shippingMethodParts) === 2) {
            $carrierProfileData = explode(':', $shippingMethodParts[1]);
            if (count($carrierProfileData) == 2) {
                return $carrierProfileData;
            }
        }

        throw new LocalizedException(__('Incorrect shipping and carrier profiles: %1', $shippingMethod));
    }

    /**
     * @param string $shippingMethod
     * @return \Bluebirdday\TranssmartSmartConnect\Model\Booking\Profile
     * @throws LocalizedException
     */
    private function getPickupProfile($shippingMethod)
    {
        list($carrierCode, $profileCode) = $this->getPickupCarrierAndProfile($shippingMethod);
        //validation
        /** @var \Bluebirdday\TranssmartSmartConnect\Model\Booking\Profile $profile */
        $profiles = $this->profileRepository->getByCodes([$profileCode]);
        $profile = array_shift($profiles);
        if (!$profile || $carrierCode !== $profile->getCarrierCode()) {
            throw new LocalizedException(__('Invalid carrier code "%1" for profile "%2"', $carrierCode, $profileCode));
        }
        return $profile;
    }

    private function getDeliveryOptionProfile($shippingMethod)
    {
        if (!$this->isDeliveryOption($shippingMethod)) {
            throw new \InvalidArgumentException('Incorrect shipping method as delivery option: ' . $shippingMethod);
        }
        $parts = explode('_', $shippingMethod);
        $profileCode = array_pop($parts);
        //validation
        /** @var \Bluebirdday\TranssmartSmartConnect\Model\Booking\Profile $profile */
        $profiles = $this->profileRepository->getByCodes([$profileCode]);
        $profile = array_shift($profiles);
        if (!$profile) {
            throw new LocalizedException(__('Profile with code "%1" not found.', $profileCode));
        }
        return $profile;
    }

    /**
     * Parse bestway method name and fetch profile from it
     *
     * @param string $shippingMethod
     * @return bool|\Bluebirdday\TranssmartSmartConnect\Model\Booking\Profile
     * @throws LocalizedException
     */
    private function getBestwayProfile($shippingMethod)
    {
        if (!$this->isBestway($shippingMethod)) {
            return false;
        }
        $methodParts = explode('_', $shippingMethod);
        $profileCode = array_pop($methodParts);

        /** @var \Bluebirdday\TranssmartSmartConnect\Model\Booking\Profile $profile */
        $profiles = $this->profileRepository->getByCodes([$profileCode]);
        $profile = array_shift($profiles);
        if (!$profile) {
            throw new LocalizedException(__('Invalid profile code "%1"', $profileCode));
        }
        return $profile;
    }

    /**
     * Parse order shipping method and get associated SmartConnect profile
     *
     * @param \Magento\Sales\Model\Order $order
     * @return false|\Bluebirdday\TranssmartSmartConnect\Model\Booking\Profile
     * @throws LocalizedException
     */
    public function parseOrder($order)
    {
        $shippingMethodsMap = $this->getShippingMethodsMap();
        $shippingMethod = $order->getShippingMethod();
        $profile = false;
        if ($this->isPickup($shippingMethod)) {
            $profile = $this->getPickupProfile($shippingMethod);
        } elseif ($this->isDeliveryOption($shippingMethod)) {
            $profile = $this->getDeliveryOptionProfile($shippingMethod);
        } elseif ($this->isBestway($shippingMethod)) {
            $profile = $this->getBestwayProfile($shippingMethod);
        } elseif ($carrierProfileCode = $this->getMappedProfile($shippingMethod, $shippingMethodsMap))  {
            $profiles = $this->profileRepository->getByCodes([$carrierProfileCode]);
            $profile = array_shift($profiles);
            if (!$profile) {
                throw new LocalizedException(__('Booking profile not found: %1', $carrierProfileCode));
            }
        }

        return $profile;
    }

    /**
     * @param $shippingMethod
     * @return bool
     */
    private function isBestway($shippingMethod)
    {
        $shippingMethodParts = explode('_', $shippingMethod);
        return $shippingMethodParts[0] === 'tablerate'
            && $shippingMethodParts[1] === 'bestway'
            && count($shippingMethodParts) === 3;
    }

    /**
     * Get pickup address of order if any
     *
     * @param \Magento\Sales\Model\Order $order
     * @param bool $reloadCollection
     * @return \Magento\Sales\Api\Data\OrderAddressInterface|null
     */
    public function getPickupAddress($order, $reloadCollection = false)
    {
        $addresses = $reloadCollection ? $order->getAddressesCollection() : $order->getAddresses();
        foreach ($addresses as $address) {
            if ($address->getAddressType() === 'pickup' && !$address->isDeleted()) {
                return $address;
            }
        }
        return null;
    }

    /**
     * Get mapped SmartConnect profile for shipping methods
     *
     * @param string $shippingMethod
     * @param \stdClass[] $map
     * @return string|null
     */
    private function getMappedProfile($shippingMethod, $map)
    {
        foreach ($map as $mapping) {
            if ($shippingMethod == $mapping->shipping_method) {
                return $mapping->carrier_profile;
            }
        }
        return null;
    }

    /**
     * Get shipping methods and mapped carrier profiles
     *
     * @return array
     * @throws LocalizedException
     */
    private function getShippingMethodsMap()
    {
        $mappedMethods = $this->scopeConfig->getValue('transsmart/shipping_methods/method_to_profile_map');
        if (!$mappedMethods) {
            return [];
        }
        $mappedMethods = json_decode($mappedMethods);

        return $mappedMethods;
    }
}
