<?php
namespace Bluebirdday\TranssmartSmartConnect\Model\Shipping;

use Magento\Framework\Exception\LocalizedException;
use Magento\Quote\Model\Quote\Address\RateRequest;
use Magento\Shipping\Model\Rate\Result;

class Pickup extends \Magento\Shipping\Model\Carrier\AbstractCarrier
    implements \Magento\Shipping\Model\Carrier\CarrierInterface
{
    /**
     * {@inheritdoc}
     */
    protected $_code = 'transsmartpickup';

    /**
     * Custom basic path for all SmartConnect Pickup method settings.
     *
     * @var string
     */
    private $baseConfigPath = 'transsmart/pickup/';

    /**
     * @var \Magento\Shipping\Model\Rate\ResultFactory
     */
    private $rateResultFactory;

    /**
     * @var \Magento\Quote\Model\Quote\Address\RateResult\MethodFactory
     */
    private $rateMethodFactory;

    /**
     * @var \Magento\Checkout\Model\Session
     */
    private $checkoutSession;

    /**
     * @var \Bluebirdday\TranssmartSmartConnect\Model\Quote\AddressManager
     */
    private $addressManager;

    /**
     * @var \Bluebirdday\TranssmartSmartConnect\Model\Booking\ProfileRepository
     */
    private $profileRepository;

    /**
     * @var \Bluebirdday\TranssmartSmartConnect\Model\CarrierRepository
     */
    private $carrierRepository;

    /**
     * @var \Magento\Quote\Model\QuoteRepository
     */
    private $quoteRepository;

    /**
     * @var \Bluebirdday\TranssmartSmartConnect\Model\Rate\Response\Parser
     */
    private $rateParser;

    /**
     * @var Method\Parser
     */
    private $methodParser;

    /**
     * Cost constructor.
     * @param \Magento\Framework\App\Config\ScopeConfigInterface $scopeConfig
     * @param \Magento\Quote\Model\Quote\Address\RateResult\ErrorFactory $rateErrorFactory
     * @param \Psr\Log\LoggerInterface $logger
     * @param \Magento\Shipping\Model\Rate\ResultFactory $rateResultFactory
     * @param \Magento\Quote\Model\Quote\Address\RateResult\MethodFactory $rateMethodFactory
     * @param \Magento\Checkout\Model\Session $checkoutSession
     * @param \Bluebirdday\TranssmartSmartConnect\Model\Rate\Response\Parser $rateParser
     * @param \Bluebirdday\TranssmartSmartConnect\Model\CarrierRepository $carrierRepository
     * @param \Bluebirdday\TranssmartSmartConnect\Model\Booking\ProfileRepository $profileRepository
     * @param \Magento\Quote\Model\QuoteRepository $quoteRepository
     * @param \Bluebirdday\TranssmartSmartConnect\Model\Quote\AddressManager $addressManager
     * @param array $data
     */
    public function __construct(
        \Magento\Framework\App\Config\ScopeConfigInterface $scopeConfig,
        \Magento\Quote\Model\Quote\Address\RateResult\ErrorFactory $rateErrorFactory,
        \Psr\Log\LoggerInterface $logger,
        \Magento\Shipping\Model\Rate\ResultFactory $rateResultFactory,
        \Magento\Quote\Model\Quote\Address\RateResult\MethodFactory $rateMethodFactory,
        \Magento\Checkout\Model\Session $checkoutSession,
        \Bluebirdday\TranssmartSmartConnect\Model\Rate\Response\Parser $rateParser,
        \Bluebirdday\TranssmartSmartConnect\Model\Shipping\Method\Parser $methodParser,
        \Bluebirdday\TranssmartSmartConnect\Model\CarrierRepository $carrierRepository,
        \Bluebirdday\TranssmartSmartConnect\Model\Booking\ProfileRepository $profileRepository,
        \Magento\Quote\Model\QuoteRepository $quoteRepository,
        \Bluebirdday\TranssmartSmartConnect\Model\Quote\AddressManager $addressManager,
        array $data = []
    ) {
        $this->carrierRepository = $carrierRepository;
        $this->profileRepository = $profileRepository;
        $this->rateResultFactory = $rateResultFactory;
        $this->rateMethodFactory = $rateMethodFactory;
        $this->checkoutSession = $checkoutSession;
        $this->quoteRepository = $quoteRepository;
        $this->addressManager = $addressManager;
        $this->rateParser = $rateParser;
        $this->methodParser = $methodParser;
        parent::__construct($scopeConfig, $rateErrorFactory, $logger, $data);
    }

    /**
     * Overridden method because of custom configuration path
     *
     * @param string $field
     * @return mixed
     */
    public function getConfigData($field)
    {
        return $this->_scopeConfig->getValue(
            $this->baseConfigPath . $field,
            \Magento\Store\Model\ScopeInterface::SCOPE_STORE,
            $this->getStore()
        );
    }

    /**
     * Overridden method because of custom configuration path
     *
     * @param string $field
     * @return bool
     */
    public function getConfigFlag($field)
    {
        return $this->_scopeConfig->isSetFlag(
            $this->baseConfigPath . 'active',
            \Magento\Store\Model\ScopeInterface::SCOPE_STORE,
            $this->getStore()
        );
    }

    /**
     * @param \Magento\Quote\Model\Quote\Address $pickupAddress
     * @return array
     */
    private function getCarrierProfile($pickupAddress)
    {
        $shippingMethod = $pickupAddress->getShippingMethod();
        $carrierAndProfile = $this->methodParser->getPickupCarrierAndProfile($shippingMethod);
        return array_pop($carrierAndProfile);
    }

    /**
     * @param RateRequest $request
     * @return bool|Result
     */
    public function collectRates(RateRequest $request)
    {
        $bookingProfile = null;

        if (!$this->getConfigFlag('active')) {
            return false;
        }

        /** @var \Magento\Shipping\Model\Rate\Result $result */
        $result = $this->rateResultFactory->create();

        /** @var \Magento\Quote\Model\Quote\Address\RateResult\Method $method */
        $method = $this->rateMethodFactory->create();

        $method->setCarrier($this->getCarrierCode());

        $quoteId = $this->checkoutSession->getQuoteId();
        if ($quoteId) {
            // when pickup address is present, get carrier and profile from it and calculate rate for this data
            //                                 set method name to carrier:profile
            // both addresses are absent use $request data for estimation
            $quote = $this->quoteRepository->getActive($quoteId);
            $address = $this->addressManager->getPickupAddress($quote)
                ?: $this->addressManager->getEstimationAddress($quote);
            if ($address) {
                $this->rateParser->updateRequestInfo($request, $address);
            }
        } else {
            // pickup shipping method is not available on backend yet
            return false;
        }

        $configuredProfiles = $this->rateParser->getPickupProfiles($request->getDestCountryId());
        if (empty($configuredProfiles)) {
            return false;
        }
        if ($address
            && $address->getShippingMethod()
            && array_key_exists($this->getCarrierProfile($address), $configuredProfiles)
        ) {
            $shippingMethod = $address->getShippingMethod();
            $carrierData = explode('_', $shippingMethod);
            if (count($carrierData) !== 2) {
                throw new \RuntimeException(__('Invalid shipping method name for pickup: %1', $shippingMethod));
            }
            $carrierData = array_pop($carrierData);
            $method->setMethod($carrierData);
            list($carrierCode, $profileCode) = explode(':', $carrierData);
            $carrier = $this->carrierRepository->loadByCode($carrierCode);
            $bookingProfile = $this->profileRepository->loadByCode($profileCode);
            $descriptionOverride = $configuredProfiles[$profileCode]->description;
            $description = $descriptionOverride ? $descriptionOverride : $bookingProfile->getDescription();
            $method->setCarrierTitle($description);
            $method->setMethodTitle($carrier->getName());
        } else {
            // this piece of code for estimates only
            $method->setCarrierTitle('');
            $method->setMethod('pickup:locations');
            $method->setMethodTitle($this->getConfigData('name'));
        }

        try {
            $amount = (float)$this->rateParser->getShippingAmount($request, $bookingProfile);
        } catch (\RuntimeException $e) {
            $amount = 0;
        } catch (LocalizedException $e) {
            $amount = 0;
        }
        $method->setPrice($amount);
        $method->setCost($amount);

        $result->append($method);

        return $result;
    }

    public function getAllowedMethods()
    {
        return ['transsmartpickup' => $this->getConfigData('name')];
    }

    public function isTrackingAvailable()
    {
        return true;
    }
}
