<?php

namespace Bluebirdday\TranssmartSmartConnect\Model\Transsmart;

use Magento\Framework\App\Config\ScopeConfigInterface;
use Magento\Framework\App\ProductMetadataInterface;
use Magento\Framework\Exception\LocalizedException;
use Magento\Framework\Module\ModuleListInterface;
use Magento\Store\Model\ScopeInterface;
use Psr\Log\LoggerInterface;

/**
 * Class Adapter
 * @package Bluebirdday\TranssmartSmartConnect\Model\SmartConnect
 */
class Adapter
{
    const ENVIRONMENT_CONFIG = 'transsmart/environment_config';

    const ENVIRONMENT_TYPE = 'transsmart/environment_config/environment';

    const MAX_ERROR_LENGTH = 50;

    /**
     * Constant is used as prefix for cache key and cache tag
     */
    const TOKEN_CACHE_KEY = 'transsmart_authorization_token';

    /**
     * @var ScopeConfigInterface
     */
    protected $scopeConfig;

    /**
     * @var LoggerInterface
     */
    protected $logger;

    /**
     * @var ProductMetadataInterface
     */
    protected $productMetadata;

    /**
     * @var ModuleListInterface
     */
    protected $moduleList;

    /**
     * @var \Magento\Integration\Model\Cache\Type
     */
    protected $cacheType;

    /**
     * @var \Magento\Config\Model\Config\Backend\Encrypted
     */
    protected $encrypted;

    /**
     * @var \Magento\Framework\DataObjectFactory
     */
    protected $dataObjectFactory;

    /**
     * @var string
     */
    protected $url;

    /**
     * @var string
     */
    protected $username;

    /**
     * @var string
     */
    protected $password;

    /**
     * Account ID also known as Client ID
     *
     * @var string
     */
    protected $accountId;

    /**
     * @var string
     */
    protected $environment;

    /**
     * Authorization token from SmartConnect. Usually expires in 24 hours
     *
     * @var string
     */
    protected $token;

    public function __construct(
        ScopeConfigInterface $scopeConfig,
        LoggerInterface $logger,
        ProductMetadataInterface $productMetadata,
        ModuleListInterface $moduleList,
        \Magento\Integration\Model\Cache\Type $cacheType,
        \Magento\Config\Model\Config\Backend\Encrypted $encrypted,
        \Magento\Framework\DataObjectFactory $dataObjectFactory
    ) {
        $this->scopeConfig = $scopeConfig;
        $this->logger = $logger;
        $this->productMetadata = $productMetadata;
        $this->moduleList = $moduleList;
        $this->cacheType = $cacheType;
        $this->encrypted = $encrypted;
        $this->dataObjectFactory = $dataObjectFactory;
    }

    /**
     * Get environment check sum
     *
     * Used for environment identification
     *
     * @return string
     */
    protected function getEnvCheckSum()
    {
        $this->initConfig();
        $environmentData = [
            $this->environment,
            $this->url,
            $this->accountId,
            $this->username,
            $this->password,
        ];
        $environmentData = implode(',', $environmentData);
        return md5($environmentData);
    }

    /**
     * Get token or read it from cache
     *
     * @return bool|string
     */
    protected function getToken()
    {
        if ($this->token) {
            return $this->token;
        }

        $checkSum = $this->getEnvCheckSum();
        $cachedValue = $this->cacheType->load(static::TOKEN_CACHE_KEY . "_$checkSum");
        if ($cachedValue) {
            $this->token = $cachedValue;
        }
        return $cachedValue;
    }

    /**
     * Save authorization token to cache
     *
     * @param string $token
     * @return bool
     */
    protected function saveCache($token)
    {
        $checkSum = $this->getEnvCheckSum();
        // save token for 24 hours(86400 seconds)
        $result = $this->cacheType->save(
            $token,
            static::TOKEN_CACHE_KEY . "_$checkSum",
            [static::TOKEN_CACHE_KEY],
            86400
        );
        if (!$result) {
            $this->logger->error('Cannot save token to cache');
        }
        return $result;
    }

    /**
     * Execute request logging it and response on debug level
     *
     * @param resource $ch
     * @param string $requestBody
     * @return mixed
     */
    protected function executeRequest($ch, $requestBody)
    {
        $time = microtime(true);
        $responseBody = curl_exec($ch);
        $time = microtime(true) - $time;

        $writer = new \Zend\Log\Writer\Stream(BP . '/var/log/transsmart.log');
        $logger = new \Zend\Log\Logger();
        $logger->addWriter($writer);
        $logger->info(sprintf(
            "REQUEST: %s\nREQUEST BODY: %s\nDURATION: %0.4f sec\nRESPONSE CODE: %s\nRESPONSE BODY: %s",
            rtrim(curl_getinfo($ch, CURLINFO_HEADER_OUT)),
            $requestBody,
            $time,
            curl_getinfo($ch, CURLINFO_HTTP_CODE),
            $responseBody
        ));

        return $responseBody;
    }

    /**
     * Get authorization token from SmartConnect
     *
     * @return bool
     * @throws LocalizedException
     */
    protected function updateToken()
    {
        $curlOptions = $this->prepareLoginOptions();

        // initialize cURL
        $ch = curl_init();
        curl_setopt_array($ch, $curlOptions);
        $responseBody = $this->executeRequest($ch, null);

        if (curl_getinfo($ch, CURLINFO_HTTP_CODE) === 200) {
            $responseObject = \json_decode($responseBody);
            if (!property_exists($responseObject, 'token') || !$responseObject->token) {
                $this->logger->emergency('Token not found in Transsmart API response');
                throw new LocalizedException(__('Token not found in Transsmart API response'));
            }
            $this->token = $responseObject->token;
            //as far as token is valid for 24 hours we save it to cache for further requests
            $this->saveCache($this->token);
            curl_close($ch);
            return true;
        }
        // collect details and close cURL resource
        $curlError = curl_error($ch) ?: $responseBody;
        curl_close($ch);
        $this->logger->critical('Cannot update authorization token: ' . $curlError);
        throw new LocalizedException(__('Cannot update authorization token'));
    }

    /**
     * @param $path
     * @param bool $required
     * @return mixed
     */
    protected function getConfigValue($path, $required = true)
    {
        if ($required && !$this->scopeConfig->isSetFlag($path)) {
            throw new \RuntimeException("Config path '$path' not found!");
        }

        return $this->scopeConfig->getValue(
            $path,
            ScopeInterface::SCOPE_STORE
        );
    }

    /**
     * Read configuration and save parameters to properties
     */
    protected function initConfig()
    {
        if ($this->environment) {
            //we imply config is already initialized
            return;
        }
        $this->environment = $this->getConfigValue(static::ENVIRONMENT_TYPE);
        $this->url = $this->getConfigValue(static::ENVIRONMENT_CONFIG . "/{$this->environment}_gateway");
        $this->username = $this->getConfigValue(static::ENVIRONMENT_CONFIG . "/username");
        $encryptedPassword = $this->getConfigValue(static::ENVIRONMENT_CONFIG . "/password");
        $this->password = $this->encrypted->processValue($encryptedPassword);
        $this->accountId = $this->getConfigValue(static::ENVIRONMENT_CONFIG . "/account_id");
    }

    /**
     * Prepare options for regular request
     *
     * @param string $method
     * @param null|string $requestBody
     * @return array
     */
    protected function prepareOptions($method, $requestBody = null)
    {
        $magentoVersion = $this->productMetadata->getVersion();
        $moduleInfo = $this->moduleList->getOne('Bluebirdday_TranssmartSmartConnect');
        $moduleVersion = is_array($moduleInfo) ? $moduleInfo['setup_version'] : null;
        if (!$this->getToken()) {
            $this->updateToken();
        }

        $curlOptions = [
            // HTTP authentication
            CURLOPT_HTTPAUTH       => CURLAUTH_BASIC,
            // HTTP request
            CURLOPT_URL            => $this->url . $method,
            CURLOPT_HTTPHEADER     => [
                'User-Agent: ' . sprintf(
                    'Magento/%s Transsmart_Shipping/%s',
                    $magentoVersion,
                    $moduleVersion
                ),
                'Authorization: Bearer ' . $this->getToken(),
                'Accept: application/json',
            ],
            // connection settings
            CURLINFO_HEADER_OUT    => true,
            CURLOPT_RETURNTRANSFER => true,
            CURLOPT_CONNECTTIMEOUT => 5,
            CURLOPT_TIMEOUT        => 60,
            CURLOPT_SSL_VERIFYHOST => 2,
            CURLOPT_SSL_VERIFYPEER => true,
        ];

        if (!is_null($requestBody)) {
            $curlOptions[CURLOPT_HTTPHEADER][] = 'Content-Type: application/json';
            $curlOptions[CURLOPT_POST]         = true;
            $curlOptions[CURLOPT_POSTFIELDS]   = $requestBody;
        }
        return $curlOptions;
    }

    /**
     * Prepare request options to get authorization token
     *
     * @return array
     */
    protected function prepareLoginOptions()
    {
        $method = '/login';
        $curlOptions = [
            // HTTP authentication
            CURLOPT_HTTPAUTH       => CURLAUTH_BASIC,
            CURLOPT_USERPWD        => $this->username . ':' . $this->password,
            // HTTP request
            CURLOPT_URL            => $this->url . $method,
            CURLOPT_HTTPHEADER     => [
                'Accept: application/json',
            ],
            // connection settings
            CURLINFO_HEADER_OUT    => true,
            CURLOPT_RETURNTRANSFER => true,
            CURLOPT_CONNECTTIMEOUT => 5,
            CURLOPT_TIMEOUT        => 20,
            CURLOPT_SSL_VERIFYHOST => 2,
            CURLOPT_SSL_VERIFYPEER => true,
        ];
        return $curlOptions;
    }

    /**
     * Wrapper for curl_init and curl_exec, which also performs error checking and logging. Returns the response body.
     *
     * @param string $method API method (e.g.: "/Api/MethodName")
     * @param string|null $requestBody This is usually a JSON encoded string
     * @return string This is usually a JSON encoded string
     * @throws LocalizedException
     */
    protected function curlExec($method, $requestBody = null)
    {
        $this->initConfig();

        $tries = 0;
        do {
            $curlOptions = $this->prepareOptions($method, $requestBody);
            // initialize cURL
            $ch = curl_init();
            curl_setopt_array($ch, $curlOptions);

            $responseBody = $this->executeRequest($ch, $requestBody);

            // collect details and close cURL resource
            $curlError = curl_error($ch);
            $httpResponseCode = curl_getinfo($ch, CURLINFO_HTTP_CODE);
            curl_close($ch);

            $tries++;
            // when we get 401, we update token, but not more than once
        } while ($httpResponseCode === 401 && $this->updateToken() && $tries < 2);

        // check for cURL failure
        if ($responseBody === false) {
            throw new LocalizedException(__('Transsmart API connection failure: %1', $curlError));
        }

        // check for unexpected HTTP response code
        if ($httpResponseCode < 200 || $httpResponseCode >= 300) {
            // try to extract JSON encoded message from response body
            $exceptionPhrase = $this->parseAndLogErrorMessage($responseBody, $method, $httpResponseCode, $requestBody);
            throw new LocalizedException($exceptionPhrase);
        }

        return $responseBody;
    }

    /**
     * Fetch error details from SmartConnect error structure.
     *
     * @param array $response
     * @return string
     */
    protected function getMessageFromArray(array $response)
    {
        $message = '';
        if (array_key_exists('message', $response)) {
            $message = $response['message'] . '.';
        }

        if (array_key_exists('description', $response)) {
            //sometimes messages from SmartConnect are too long ;)
            $message .= ' ' . substr($response['description'], 0, static::MAX_ERROR_LENGTH);
        }

        if (array_key_exists('details', $response) && is_array($response['details'])) {
            $firstDetail = reset($response['details']);
            // we hope SmartConnect will fix this typo
            if (is_array($firstDetail) && array_key_exists('errorDecription', $firstDetail)) {
                $message .= ' ' . $firstDetail['errorDecription'];
            }
            // this part without typo will work when SmartConnect is ready
            if (is_array($firstDetail) && array_key_exists('errorDescription', $firstDetail)) {
                $message .= ' ' . $firstDetail['errorDescription'];
            }
        }

        return $message;
    }

    /**
     * @param string|null $responseBody
     * @param string $method
     * @param int $httpResponseCode
     * @param null|string $requestBody
     * @return \Magento\Framework\Phrase
     */
    protected function parseAndLogErrorMessage($responseBody, $method, $httpResponseCode, $requestBody)
    {
        $message = false;
        if ($responseBody !== false) {
            try {
                $response = \Zend_Json_Decoder::decode($responseBody);
                $this->logger->critical(
                    sprintf(
                        'Transsmart error for URL "%s"; request body: %s; response body: %s',
                        $method,
                        var_export($requestBody, 1),
                        var_export($response, 1)
                    )
                );
                if (!is_array($response)) {
                    $message = substr((string)$response, 0, static::MAX_ERROR_LENGTH);
                } else {
                    $message = $this->getMessageFromArray($response);
                }
            } catch (\Zend_Json_Exception $exception) {
                $this->logger->critical(
                    sprintf('Transsmart URL %s decoding exception: %s', $method, $exception->getMessage())
                );
                return __('Cannot decode Transsmart error');
            }
        }

        if ($message) {
            return __('Transsmart API error: %1', $message);
        }

        return __('Transsmart API returned unexpected HTTP response code: %1', $httpResponseCode);
    }

    /**
     * Common method for list setting retrieval
     *
     * @param string $entity
     * @return array
     */
    protected function getListSettings($entity)
    {
        $this->initConfig();
        $url = "/v2/accounts/{$this->accountId}/listsettings/$entity";
        $items = json_decode($this->curlExec($url));
        if (!is_array($items)) {
            throw new \RuntimeException("Unexpected response for URl '$url': $items");
        }

        $decodedItems = [];
        foreach ($items as $item) {
            $decodedItems[] = json_decode($item->value);
        }

        return $decodedItems;
    }

    /**
     * Request pickup locations around the defined address
     *
     * @param \Bluebirdday\TranssmartSmartConnect\Model\Location\Request $locationsRequest
     * @return object|null
     */
    public function getPickupLocations(\Bluebirdday\TranssmartSmartConnect\Model\Location\Request $locationsRequest)
    {
        $this->initConfig();
        $url = "/v2/locations/{$this->accountId}/?$locationsRequest";
        $items = json_decode($this->curlExec($url));
        if (!is_object($items)) {
            return null;
        }
        if (!property_exists($items, 'sources') || !is_array($items->sources)) {
            throw new \RuntimeException("Unexpected response for URL '$url'");
        }

        return $items;
    }

    /**
     * Book or Book-and-Print shipments
     *
     * @see https://devdocs.transsmart.com/#_shipment_booking
     * @param \Bluebirdday\TranssmartSmartConnect\Model\Data\Shipment[] $shipments
     * @param bool $print
     * @return mixed
     */
    public function bookShipments($shipments, $print = false)
    {
        $shipmentsData = [];
        $this->initConfig();
        $action = $print ? 'PRINT' : 'BOOK';
        $url = "/v2/shipments/{$this->accountId}/$action?rawJob=false";
        foreach ($shipments as $shipment) {
            $shipmentsData[] = $shipment->toArray();
        }
        $requestBody = json_encode($shipmentsData);
        $response = json_decode($this->curlExec($url, $requestBody));

        return $response;
    }

    /**
     * Request shipment status update
     *
     * @see https://devdocs.transsmart.com/#_shipment_status_tracking
     * @param string $shipmentId
     * @param bool $detailed
     * @param bool $currentStatusOnly
     * @return mixed
     */
    public function updateShipmentStatus($shipmentId, $detailed = true, $currentStatusOnly = false)
    {
        $this->initConfig();

        $url = "/v2/statuses/{$this->accountId}/shipments/$shipmentId?isDetailed=$detailed"
            . "&currentStatusOnly=$currentStatusOnly";
        $response = json_decode($this->curlExec($url));

        return $response;
    }

    /**
     * Get and update shipments api call.
     *
     * @see https://devdocs.transsmart.com/#_shipment_status_tracking
     * @param array $params
     * @return bool|\Magento\Framework\DataObject
     */
    public function updateShipmentStatuses(array $params)
    {
        // dateFrom, dateTo, maxResults are required params
        // both dates have to be in 'YYYY-MM-dd HH:mm:ss' format
        $this->initConfig();
        $url = "/v2/statuses/{$this->accountId}/shipments?" . http_build_query($params);
        $response = json_decode($this->curlExec($url), true);
        if (isset($response['content'])) {
            $values = $response['content'];
            $keys = array_column($values, 'reference');

            return $this->dataObjectFactory->create([
                'data' => array_combine($keys, $values)
            ]);
        }

        return false;
    }

    /**
     * Print document with SmartPrint
     *
     * @see https://devdocs.transsmart.com/#_print_document
     * @param string $reference SmartConnect shipment reference(shipment increment id)
     * @return bool
     */
    public function printShipment($reference)
    {
        $this->initConfig();
        $url = "/v2/prints/{$this->accountId}/$reference?rawJob=false";
        $response = $this->curlExec($url);
        return $response === 'Sent to SmartPrint';
    }

    /**
     * Get shipping rates
     *
     * @see https://devdocs.transsmart.com/#_calculating_rates
     * @param \Bluebirdday\TranssmartSmartConnect\Model\Data\Rate $requestData
     * @return array|null
     */
    public function getRates(\Bluebirdday\TranssmartSmartConnect\Model\Data\Rate $requestData)
    {
        $this->initConfig();
        $url = "/v2/rates/{$this->accountId}?action=false&onlyrates=false";
        $requestDataMod = [$requestData->toArray()];
        $requestBody = json_encode($requestDataMod);
        $response = json_decode($this->curlExec($url, $requestBody));
        if (empty($response) || !is_array($response)) {
            return null;
        }
        $firstItem = array_shift($response);
        if (!array_key_exists('rates', $firstItem)) {
            return null;
        }
        return $firstItem->rates;
    }

    /**
     * Get configured carrier profiles.
     *
     * @return array
     * @throws \RuntimeException
     */
    public function getCarriers()
    {
        return $this->getListSettings('carriers');
    }

    /**
     * Get packages
     *
     * @return array
     * @throws \RuntimeException
     */
    public function getPackages()
    {
        return $this->getListSettings('packages');
    }

    /**
     * Get service level times
     *
     * @return array
     * @throws \RuntimeException
     */
    public function getServiceLevelTimes()
    {
        return $this->getListSettings('serviceLevelTimes');
    }

    /**
     * Get other service levels
     *
     * @return array
     * @throws \RuntimeException
     */
    public function getServiceLevelOthers()
    {
        return $this->getListSettings('serviceLevelOthers');
    }

    /**
     * Get booking profiles
     *
     * @return array
     * @throws \RuntimeException
     */
    public function getBookingProfiles()
    {
        return $this->getListSettings('bookingProfiles');
    }

    /**
     * Get mail types list
     *
     * @return array
     */
    public function getMailTypes()
    {
        return $this->getListSettings('mailTypes');
    }

    /**
     * Get Incoterms list
     *
     * @return array
     */
    public function getIncoterms()
    {
        return $this->getListSettings('incoterms');
    }

    /**
     * Get Cost Centers list
     *
     * @return array
     */
    public function getCostCenters()
    {
        return $this->getListSettings('costCenters');
    }
}
